// This file is part of Gopher2600.
//
// Gopher2600 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// Gopher2600 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with Gopher2600.  If not, see <https://www.gnu.org/licenses/>.

package sdlimgui

import (
	"github.com/jetsetilly/gopher2600/curated"
	"github.com/jetsetilly/gopher2600/emulation"
	"github.com/jetsetilly/gopher2600/gui"
	"github.com/jetsetilly/gopher2600/hardware/memory/cartridge/mapper"
	"github.com/jetsetilly/gopher2600/hardware/riot/ports/plugging"
)

type featureRequest struct {
	request gui.FeatureReq
	args    []gui.FeatureReqData
}

// GetFeature implements gui.GUI interface.
func (img *SdlImgui) GetFeature(request gui.FeatureReq) (gui.FeatureReqData, error) {
	img.polling.featureGet <- featureRequest{request: request}
	return <-img.polling.featureGetData, <-img.polling.featureGetErr
}

// featureRequests have been handed over to the featureGet channel. we service
// any requests on that channel here.
func (img *SdlImgui) serviceGetFeature(request featureRequest) {
	switch request.request {
	case gui.ReqState:
		img.polling.featureGetData <- img.state
		img.polling.featureGetErr <- nil
	default:
		img.polling.featureGetData <- nil
		img.polling.featureGetErr <- curated.Errorf(gui.UnsupportedGuiFeature, request.request)
	}
}

// SetFeature implements gui.GUI interface.
func (img *SdlImgui) SetFeature(request gui.FeatureReq, args ...gui.FeatureReqData) error {
	img.polling.featureSet <- featureRequest{request: request, args: args}
	return <-img.polling.featureSetErr
}

// SetFeatureNoError implements gui.GUI interface.
func (img *SdlImgui) SetFeatureNoError(request gui.FeatureReq, args ...gui.FeatureReqData) {
	img.polling.featureSet <- featureRequest{request: request, args: args}
	go func() {
		<-img.polling.featureSetErr
	}()
}

// check length of arguments sent with feature request.
func argLen(args []gui.FeatureReqData, expectedLen int) error {
	if len(args) != expectedLen {
		return curated.Errorf("wrong number of arguments (%d instead of %d)", len(args), expectedLen)
	}
	return nil
}

// featureRequests have been handed over to the featureReq channel. we service
// any requests on that channel here.
func (img *SdlImgui) serviceSetFeature(request featureRequest) {
	var err error

	switch request.request {
	case gui.ReqSetEmulation:
		err = argLen(request.args, 1)
		if err == nil {
			img.setEmulation(request.args[0].(emulation.Emulation))
		}

	case gui.ReqState:
		err = argLen(request.args, 1)
		if err == nil {
			img.setEmulationState(request.args[0].(emulation.State))
		}

	case gui.ReqMonitorSync:
		err = argLen(request.args, 1)
		if err == nil {
			img.screen.crit.section.Lock()
			img.screen.crit.monitorSync = request.args[0].(bool)
			img.screen.crit.section.Unlock()
		}

	case gui.ReqFullScreen:
		err = argLen(request.args, 1)
		if err == nil {
			img.plt.setFullScreen(request.args[0].(bool))
		}

	case gui.ReqSetVisibility:
		err = argLen(request.args, 1)
		if err == nil {
			if img.isPlaymode() {
				err = curated.Errorf("visibility not supported in playmode")
			} else {
				img.wm.dbgScr.setOpen(request.args[0].(bool))
			}
		}

	case gui.ReqPlusROMFirstInstallation:
		err = argLen(request.args, 1)
		if err == nil {
			img.plusROMFirstInstallation = request.args[0].(*gui.PlusROMFirstInstallation)
		}

	case gui.ReqControllerChange:
		if img.state == emulation.Initialising {
			break
		}

		if img.isPlaymode() {
			port := request.args[0].(plugging.PortID)
			switch port {
			case plugging.PortLeftPlayer:
				img.playScr.peripheralLeft.set(request.args[1].(plugging.PeripheralID))
			case plugging.PortRightPlayer:
				img.playScr.peripheralRight.set(request.args[1].(plugging.PeripheralID))
			}
		}

	case gui.ReqCartridgeEvent:
		if img.state == emulation.Initialising {
			break
		}

		if img.isPlaymode() {
			img.playScr.cartridgeEvent.set(request.args[0].(mapper.Event))
		}

	default:
		err = curated.Errorf(gui.UnsupportedGuiFeature, request.request)
	}

	if err == nil {
		img.polling.featureSetErr <- nil
	} else {
		img.polling.featureSetErr <- curated.Errorf("sdlimgui: %s: %v", request.request, err)
	}
}
